import dotenv
import io
import logging
import os
import pprint
import shutil
import subprocess

from abb.jsonld import set_jsonld_context
from tornado.ioloop import IOLoop
from tornado.options import parse_command_line
from tornado.gen import coroutine, maybe_future
from unicodedata import normalize


class ExecutionFailed(RuntimeError):
    pass


class ChangeWorkingDir(object):
    def __init__(self, new_dir):
        self.new_dir = new_dir

    def __enter__(self):
        self.working_dir = os.getcwd()
        logging.debug("Changing directory: %s -> %s", self.working_dir, self.new_dir)
        os.chdir(self.new_dir)
        return self

    def __exit__(self, *args):
        logging.debug("Changing directory back: %s", self.working_dir)
        os.chdir(self.working_dir)


def rmdirs(*dirs):
    for path in dirs:
        logging.debug("Removing directory: %s", path)
        shutil.rmtree(path, ignore_errors=True)


def slugify(name, replace_old=" ", replace_new="_"):
    if not name:
        return ""
    name = normalize('NFKD', name)
    name = name.encode('ascii', 'ignore')
    name = name.decode("ascii")
    name = name.replace(replace_old, replace_new)
    return name


def logprint(obj, name=""):
    out = io.StringIO()
    pprint.pprint(obj, stream=out, indent=2)
    value = out.getvalue().rstrip()
    if "\n" in value:
        value = "\n%s" % value
    text = "logprint(%s): %s" % (name, value)
    return text


def execute(arguments, expect=0, raw=False):
    logging.debug("Executing: %s" % " ".join(arguments))
    p = subprocess.Popen(arguments, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    result, errors = p.communicate()

    if errors:
        errors = errors.decode("utf-8")
        for line in errors.splitlines():
            logging.warning("Execution stderr: %s", line)

    if expect is not None:
        if isinstance(expect, int):
            expect = [expect]
        if p.returncode not in expect:
            message = "Command returns an unexpected code: %s" % p.returncode
            raise ExecutionFailed(message)

    if not raw:
        result = result.decode("utf-8")

    return result


def bootstrap(callback, *callback_args):
    envfile = os.path.join(os.getcwd(), '.env')
    dotenv.read_dotenv(envfile)
    argv = parse_command_line()
    set_jsonld_context()

    @coroutine
    def runner():
        args = (argv,) + callback_args
        if not callback.__closure__:
            call = coroutine(callback)
        else:
            call = callback
        result = yield maybe_future(call(*args))
        return result

    if callback is not None:
        return IOLoop.instance().run_sync(runner)

    return True
